@@ -232,9 +232,6 @@ custom_vmap_p.def_abstract_eval(custom_vmap_abstract_eval)
 batching.primitive_batchers[custom_vmap_p] = custom_vmap_batching
 ad.primitive_jvps[custom_vmap_p] = custom_vmap_jvp
 xla.register_initial_style_primitive(custom_vmap_p)
-xla.register_translation(custom_vmap_p,
-                         xla.lower_fun(custom_vmap_impl, new_style=True,
-                                       multiple_results=True))
 mlir.register_lowering(custom_vmap_p, mlir.lower_fun(
     custom_vmap_impl, multiple_results=True))
 
