@@ -22,8 +22,9 @@ from typing import (Callable, Generic, Optional, Sequence, Tuple, List, TypeVar,
 from jax import core
 from jax import linear_util as lu
 from jax.custom_transpose import custom_transpose
-from jax.tree_util import (tree_flatten, tree_unflatten, tree_map, treedef_is_leaf,
-                           treedef_tuple, register_pytree_node_class)
+from jax.tree_util import (tree_flatten, tree_unflatten, tree_map,
+                           treedef_is_leaf, treedef_tuple,
+                           register_pytree_node_class, tree_leaves)
 from jax._src import custom_api_util
 from jax._src import dtypes
 from jax._src.util import cache, safe_zip, safe_map, split_list, Unhashable
@@ -398,10 +399,6 @@ def _custom_jvp_call_jaxpr_vmap(
 batching.axis_primitive_batchers[custom_jvp_call_jaxpr_p] = _custom_jvp_call_jaxpr_vmap
 
 xla.register_initial_style_primitive(custom_jvp_call_jaxpr_p)
-xla.register_translation(
-    custom_jvp_call_jaxpr_p,
-    xla.lower_fun(_custom_jvp_call_jaxpr_impl, new_style=True,
-                  multiple_results=True))
 
 # If a (multi)linear function is defined with a custom jvp, then
 # custom_jvp_call_jaxpr can appear in jaxprs to be transposed. Since it's
@@ -568,21 +565,17 @@ class custom_vjp(Generic[ReturnValue]):
       out_tree = aux if fst else aux[0]
       return tree_unflatten(out_tree, out_flat)
 
-@partial(partial, tree_map)
 def _check_for_tracers(x):
-  if isinstance(x, core.Tracer):
-    msg = ("Found a JAX Tracer object passed as an argument to a custom_vjp "
-           "function in a position indicated by nondiff_argnums as "
-           "non-differentiable. Tracers cannot be passed as non-differentiable "
-           "arguments to custom_vjp functions; instead, nondiff_argnums should "
-           "only be used for arguments that can't be or contain JAX tracers, "
-           "e.g. function-valued arguments. In particular, array-valued "
-           "arguments should typically not be indicated as nondiff_argnums. "
-           "\n\n"
-           "This behavior recently changed in JAX. "
-           "See https://github.com/google/jax/blob/main/docs/custom_vjp_update.md "
-           "for more information.")
-    raise UnexpectedTracerError(msg)
+  for leaf in tree_leaves(x):
+    if isinstance(x, core.Tracer):
+      msg = ("Found a JAX Tracer object passed as an argument to a custom_vjp "
+            "function in a position indicated by nondiff_argnums as "
+            "non-differentiable. Tracers cannot be passed as non-differentiable "
+            "arguments to custom_vjp functions; instead, nondiff_argnums should "
+            "only be used for arguments that can't be or contain JAX tracers, "
+            "e.g. function-valued arguments. In particular, array-valued "
+            "arguments should typically not be indicated as nondiff_argnums.")
+      raise UnexpectedTracerError(msg)
 
 @lu.transformation_with_aux
 def _flatten_fwd(in_tree, *args):
@@ -614,11 +607,11 @@ def _flatten_bwd(in_tree, in_avals, out_trees, *args):
   zero = object()  # non-pytree sentinel to replace Nones in py_cts_in
   dummy = tree_unflatten(in_tree, [object()] * in_tree.num_leaves)
   cts_in_flat = []
-  append_cts = lambda x, d: cts_in_flat.extend([x] * len(tree_flatten(d)[0]))
+  append = lambda x, d: cts_in_flat.extend([x] * len(tree_flatten(d)[0])) or x
   try:
     if not isinstance(py_cts_in, tuple):
       raise ValueError
-    tree_map(append_cts,
+    tree_map(append,
              tuple(zero if ct is None else ct for ct in py_cts_in), dummy)
   except ValueError:
     _, in_tree2 = tree_flatten(py_cts_in)
@@ -769,13 +762,9 @@ def _custom_vjp_call_jaxpr_vmap(
 batching.axis_primitive_batchers[custom_vjp_call_jaxpr_p] = _custom_vjp_call_jaxpr_vmap
 
 xla.register_initial_style_primitive(custom_vjp_call_jaxpr_p)
-xla.register_translation(
-    custom_vjp_call_jaxpr_p,
-    xla.lower_fun(_custom_vjp_call_jaxpr_impl, new_style=True,
-                  multiple_results=True))
 
 batching.primitive_batchers[ad.custom_lin_p] = ad._raise_custom_vjp_error_on_jvp
-xla.register_translation(ad.custom_lin_p, ad._raise_custom_vjp_error_on_jvp)
+mlir.register_lowering(ad.custom_lin_p, ad._raise_custom_vjp_error_on_jvp)
 
 pe.partial_eval_jaxpr_custom_rules[custom_vjp_call_jaxpr_p] = \
     custom_jvp_jaxpr_custom_partial_eval_rule  # type: ignore
@@ -1165,9 +1154,6 @@ linear_call_p.def_impl(_linear_call_impl)
 linear_call_p.def_abstract_eval(_linear_call_abstract_eval)
 ad.primitive_transposes[linear_call_p] = _linear_call_transpose_rule
 xla.register_initial_style_primitive(linear_call_p)
-xla.register_translation(linear_call_p,
-                         xla.lower_fun(_linear_call_impl, new_style=True,
-                                       multiple_results=True))
 mlir.register_lowering(linear_call_p, mlir.lower_fun(
     _linear_call_impl, multiple_results=True))
 
@@ -1185,9 +1171,8 @@ unreachable_p.def_impl(unreachable_impl)
 
 # Translation raises an exception
 # TODO(frostig,mattjj): We have no good way to translate a function
-# that errs. Since translation over-approximates concrete evaluation,
-# we err on translation for the time being.
-xla.register_translation(unreachable_p, unreachable_impl)
+# that errs. Since MHLO lowering over-approximates concrete evaluation,
+# we err on MHLO lowering for the time being.
 mlir.register_lowering(unreachable_p, unreachable_impl)
 
 # Abstract evaluation proceeds without issue, to allow for staging
