@@ -194,7 +194,7 @@ xla.xla_call_p.def_impl(_xla_call_impl)
 
 def _xla_callable_uncached(fun: lu.WrappedFun, device, backend, name,
                            donated_invars, *arg_specs):
-  return lower_xla_callable(fun, device, backend, name, donated_invars,
+  return lower_xla_callable(fun, device, backend, name, donated_invars, False,
                             *arg_specs).compile().unsafe_call
 
 _xla_callable = lu.cache(_xla_callable_uncached)
@@ -214,7 +214,7 @@ def log_elapsed_time(fmt: str):
 
 @profiler.annotate_function
 def lower_xla_callable(fun: lu.WrappedFun, device, backend, name,
-                       donated_invars, *arg_specs):
+                       donated_invars, always_lower: bool, *arg_specs):
   if device is not None and backend is not None:
     raise ValueError("can't specify both a device and a backend for jit, "
                      "got device={} and backend={}".format(device, backend))
@@ -227,8 +227,6 @@ def lower_xla_callable(fun: lu.WrappedFun, device, backend, name,
                         "for jit in {elapsed_time} sec"):
     jaxpr, out_avals, consts = pe.trace_to_jaxpr_final(
         fun, abstract_args, pe.debug_info_final(fun, "jit"), which_explicit)
-  if jaxpr.effects:
-    raise NotImplementedError('Lowering jaxprs with effects not supported.')
   if any(isinstance(c, core.Tracer) for c in consts):
     raise UnexpectedTracerError("Encountered an unexpected tracer.")
   # TODO(mattjj): handle argument pruning w/ dynamic shapes
@@ -255,7 +253,7 @@ def lower_xla_callable(fun: lu.WrappedFun, device, backend, name,
   # Computations that only produce constants and/or only rearrange their inputs,
   # which are often produced from partial evaluation, don't need compilation,
   # and don't need to evaluate their arguments.
-  if not jaxpr.eqns:
+  if not jaxpr.eqns and not always_lower:
     return XlaComputation(
         name, None, True, None, None, jaxpr=jaxpr, consts=consts, device=device,
         in_avals=abstract_args, out_avals=out_avals, kept_var_idx=kept_var_idx)
@@ -290,7 +288,7 @@ def lower_xla_callable(fun: lu.WrappedFun, device, backend, name,
   # pass long arg lists as tuple for TPU
   tuple_args = len(abstract_args) > 100
   axis_env = xla.AxisEnv(nreps, (), ())
-  name_stack = xla.new_name_stack(xla.wrap_name(name, 'jit'))
+  name_stack = util.new_name_stack(util.wrap_name(name, 'jit'))
   closed_jaxpr = core.ClosedJaxpr(jaxpr, consts)
   module_name = f"jit_{fun.__name__}"
   module = mlir.lower_jaxpr_to_module(
@@ -672,7 +670,11 @@ def compile_or_get_cached(backend, computation, compile_options):
   if isinstance(computation, ir.Module):
     sym_name = computation.operation.attributes['sym_name']
     module_name = ir.StringAttr(sym_name).value
-    computation = mlir.module_to_string(computation)
+    # Convert ir.Module to str representation (the default), unless the
+    # back-end expliclity flags the ability to handle a module directly
+    # (avoiding the overhead of back and forth conversions)
+    if getattr(backend, "needs_str_ir", True):
+      computation = mlir.module_to_string(computation)
   else:
     module_name = computation.name()
 
@@ -690,8 +692,13 @@ def compile_or_get_cached(backend, computation, compile_options):
       return compiled
 
   if FLAGS.jax_dump_ir_to:
-    ir_str = (computation if isinstance(computation, str)
-              else computation.as_hlo_text())
+    if isinstance(computation, xc.XlaComputation):
+      ir_str = computation.as_hlo_text()
+    elif isinstance(computation, ir.Module):
+      ir_str = mlir.module_to_string(computation)
+    else:
+      assert isinstance(computation, str)
+      ir_str = computation
     _dump_ir_to_file(module_name, ir_str)
   return backend_compile(backend, computation, compile_options)
 
@@ -860,7 +867,6 @@ def _device_put_impl(x, device: Optional[Device] = None):
 device_put_p = core.Primitive('device_put')
 device_put_p.def_impl(_device_put_impl)
 device_put_p.def_abstract_eval(lambda x, device=None: x)
-xla.translations[device_put_p] = lambda c, x, device=None: x
 ad.deflinear2(device_put_p, lambda cotangent, _, **kwargs: [cotangent])
 masking.defvectorized(device_put_p)
 batching.defvectorized(device_put_p)
