@@ -27,20 +27,24 @@ def initialize(coordinator_address: str, num_processes: int, process_id: int):
   is not required for CPU or TPU backends.
 
   Args:
-    coordinator_address: IP address of the coordinator.
+    coordinator_address: IP address and port of the coordinator. The choice of
+      port does not matter, so long as the port is available on the coordinator
+      and all processes agree on the port.
     num_processes: Number of processes.
-    process_id: Id of the current processe.
+    process_id: Id of the current process.
 
   Example:
 
   Suppose there are two GPU hosts, and host 0 is the designated coordinator
-  with address '10.0.0.1:1234', to initialize the GPU cluster, run the
+  with address ``10.0.0.1:1234``. To initialize the GPU cluster, run the
   following commands before anything else.
 
-  On host 0
+  On host 0:
+
   >>> jax.distributed.initialize('10.0.0.1:1234', 2, 0)  # doctest: +SKIP
 
-  On host 1
+  On host 1:
+
   >>> jax.distributed.initialize('10.0.0.1:1234', 2, 1)  # doctest: +SKIP
   """
   if process_id == 0:
