@@ -47,7 +47,6 @@ except (ModuleNotFoundError, ImportError):
 traceback_util.register_exclusion(__file__)
 
 
-xops = xla_client.ops
 XlaBackend = xla_client._xla.Client
 
 FLAGS = flags.FLAGS
@@ -113,9 +112,8 @@ def get_compile_options(
   compile_options.num_partitions = num_partitions
   build_options = compile_options.executable_build_options
   build_options.use_spmd_partitioning = use_spmd_partitioning
-  if jax._src.lib.xla_extension_version >= 61:
-    build_options.use_auto_spmd_partitioning = use_auto_spmd_partitioning
-  if use_auto_spmd_partitioning and jax._src.lib.xla_extension_version >= 62:
+  build_options.use_auto_spmd_partitioning = use_auto_spmd_partitioning
+  if use_auto_spmd_partitioning:
     build_options.auto_spmd_partitioning_mesh_shape = auto_spmd_partitioning_mesh_shape
     build_options.auto_spmd_partitioning_mesh_ids = auto_spmd_partitioning_mesh_ids
   if device_assignment is not None:
@@ -222,6 +220,11 @@ register_backend_factory(
 if iree is not None:
   register_backend_factory("iree", iree.iree_client_factory, priority=-100)
 
+def is_known_platform(platform: str):
+  # A platform is valid if there is a registered factory for it. It does not
+  # matter if we were unable to initialize that platform; we only care that
+  # we've heard of it and it isn't, e.g., a typo.
+  return platform in _backend_factories.keys()
 
 def backends():
   global _backends
