@@ -59,6 +59,8 @@ Effects = Set[Effect]
 
 no_effects: Effects = set()
 
+ordered_effects: Set[Effect] = set()
+
 class Jaxpr:
   constvars: List[Var]
   invars: List[Var]
@@ -478,6 +480,25 @@ class Trace:
            "to handle custom_vjp primitives")
     raise NotImplementedError(msg)
 
+
+def raise_as_much_as_possible(tracer) -> Tracer:
+  # Find effective bottom of trace stack (highest dynamic Trace on the stack).
+  trace_stack = thread_local_state.trace_state.trace_stack.stack
+  idx = next(i for i, m in enumerate(trace_stack) if m is
+             thread_local_state.trace_state.trace_stack.dynamic)
+
+  # Only pay attention to effective part of trace stack.
+  trace_stack = trace_stack[idx:]
+
+  # Lift tracer into everything in the effective stack higher than its level
+  for trace in trace_stack:
+    trace = trace.with_cur_sublevel()
+    if (not isinstance(tracer, Tracer) or tracer._trace.level < trace.level):
+      tracer = trace.full_raise(tracer)
+
+  return tracer
+
+
 def escaped_tracer_error(tracer, detail=None):
   num_frames = FLAGS.jax_tracer_error_num_traceback_frames
   msg = ('Encountered an unexpected tracer. A function transformed by JAX '
@@ -1511,7 +1532,11 @@ class DimensionHandler:
     return sz1 // sz2
 
   def stride(self, d: DimSize, window_size: DimSize, window_stride: DimSize) -> DimSize:
-    """(d - window_size) // window_stride + 1"""
+    """(d - window_size) // window_stride + 1.
+
+    If d == 0 or window_size > d, returns 0.
+    """
+    if d == 0 or window_size > d: return 0
     return (d - window_size) // window_stride + 1
 
   def dilate(self, d: DimSize, dilation: int) -> DimSize:
@@ -2292,7 +2317,7 @@ def check_call(ctx_factory, prim, in_avals, params):
 
   # These checks also happen in recursive call, but give better errors here.
   if len(in_avals) != len(call_jaxpr.invars):
-    raise JaxprTypeError(f"Call primitive {prim} with {len(call_jaxpr.invars)} "
+    raise JaxprTypeError(f"Call primitive {prim} with {len(in_avals)} "
                          f"operands cannot call jaxpr with {len(call_jaxpr.invars)} "
                          f"inputs")
   binder_avals = [v.aval for v in call_jaxpr.invars]
