@@ -182,7 +182,6 @@ To fit the same model on sparse data, we can apply the :func:`sparsify` transfor
      -0.670236    0.03132951 -0.05356663]
 """
 
-# flake8: noqa: F401
 from jax.experimental.sparse.ad import (
     grad as grad,
     value_and_grad as value_and_grad,
@@ -190,6 +189,7 @@ from jax.experimental.sparse.ad import (
 from jax.experimental.sparse.bcoo import (
     bcoo_add_batch_dim as bcoo_add_batch_dim,
     bcoo_broadcast_in_dim as bcoo_broadcast_in_dim,
+    bcoo_concatenate as bcoo_concatenate,
     bcoo_dot_general as bcoo_dot_general,
     bcoo_dot_general_p as bcoo_dot_general_p,
     bcoo_dot_general_sampled as bcoo_dot_general_sampled,
@@ -201,7 +201,11 @@ from jax.experimental.sparse.bcoo import (
     bcoo_multiply_dense as bcoo_multiply_dense,
     bcoo_multiply_sparse as bcoo_multiply_sparse,
     bcoo_reduce_sum as bcoo_reduce_sum,
+    bcoo_sort_indices as bcoo_sort_indices,
+    bcoo_sort_indices_p as bcoo_sort_indices_p,
     bcoo_spdot_general_p as bcoo_spdot_general_p,
+    bcoo_sum_duplicates as bcoo_sum_duplicates,
+    bcoo_sum_duplicates_p as bcoo_sum_duplicates_p,
     bcoo_todense as bcoo_todense,
     bcoo_todense_p as bcoo_todense_p,
     bcoo_transpose as bcoo_transpose,
