@@ -14,6 +14,7 @@
 
 # Lowering and execution path that converts jaxprs into the MLIR MHLO/CHLO
 # dialects.
+from __future__ import annotations
 
 import collections
 import dataclasses
@@ -38,6 +39,7 @@ from jax._src.lib.mlir import ir
 from jax._src.lib.mlir.dialects import chlo
 from jax._src.lib.mlir.dialects import mhlo
 from jax._src.lib.mlir.dialects import func as func_dialect
+from jax._src.lib import xla_bridge as xb
 from jax._src.lib import xla_client as xc
 from jax._src import source_info_util
 import jax._src.util as util
@@ -47,16 +49,6 @@ import jax.interpreters.partial_eval as pe
 import jax.interpreters.xla as xla
 import numpy as np
 
-# TODO(jakevdp): remove this when minimum_jaxlib_version >= 0.3.3
-if jax._src.lib.mlir_api_version >= 4:
-  FuncOp = func_dialect.FuncOp  # pytype: disable=module-attr
-else:
-  from jax._src.lib.mlir.dialects import builtin  # pylint: disable=import-not-at-top
-  FuncOp = builtin.FuncOp  # pytype: disable=module-attr
-# mypy gets confused by conditional imports, so alias to Any for now.
-FuncOpType = Any
-
-
 map, unsafe_map = util.safe_map, map
 zip, unsafe_zip = util.safe_zip, zip
 
@@ -65,6 +57,8 @@ T = typing.TypeVar("T")
 # mypy implicitly sets this variable to true when type checking.
 MYPY = False
 
+lowerable_effects: Set[core.Effect] = set()
+
 
 # IR Helpers
 
@@ -201,7 +195,7 @@ def ir_constants(val: Any,
 
 def ir_constant(val: Any, canonicalize_types: bool = True) -> ir.Value:
   """Convenience wrapper around ir_constants for singleton values."""
-  values = ir_constants(val)
+  values = ir_constants(val, canonicalize_types=canonicalize_types)
   if len(values) != 1:
     raise TypeError(f"ir_constant called on {val} which corresponds to "
                     f"multiple IR values {values}")
@@ -373,7 +367,7 @@ class ModuleContext:
   name_stack: NameStack
 
   # Cached primitive lowerings.
-  cached_primitive_lowerings: Dict[Any, FuncOpType]
+  cached_primitive_lowerings: Dict[Any, func_dialect.FuncOp]
 
   @property
   def axis_env(self) -> xla.AxisEnv:
@@ -388,7 +382,7 @@ class ModuleContext:
       module: Optional[ir.Module] = None,
       ip: Optional[ir.InsertionPoint] = None,
       symbol_table: Optional[ir.SymbolTable] = None,
-      cached_primitive_lowerings: Optional[Dict[Any, FuncOpType]] = None):
+      cached_primitive_lowerings: Optional[Dict[Any, func_dialect.FuncOp]] = None):
     assert platform is not None
     self.context = context or make_ir_context()
     self.module = module or ir.Module.create(loc=ir.Location.unknown(self.context))
@@ -410,6 +404,12 @@ class LoweringRuleContext:
   primitive: Optional[core.Primitive]
   avals_in: Sequence[core.AbstractValue]
   avals_out: Any  # Usually Sequence[core.AbstractValue], but sometimes None.
+  tokens_in: TokenSet
+  tokens_out: Optional[TokenSet]  # Mutable store for output containers
+
+  def set_tokens_out(self, tokens_out: TokenSet):
+    assert self.tokens_out is None, 'Should only set `tokens_out` once.'
+    self.tokens_out = tokens_out
 
   def replace(self, **kw): return dataclasses.replace(self, **kw)
 
@@ -485,6 +485,8 @@ def lower_jaxpr_to_module(
 
   Handles the quirks of the argument/return value passing conventions of the
   runtime."""
+  if not xb.is_known_platform(platform):
+    raise ValueError(f"Unknown platform {platform}")
   input_output_aliases = None
   in_avals = jaxpr.in_avals
   if arg_shardings is not None:
@@ -521,8 +523,15 @@ def lower_jaxpr_to_module(
     ctx.module.operation.attributes["sym_name"] = ir.StringAttr.get(
         f"{module_name}.{next(_module_unique_id)}")
     # TODO(phawkins): represent units with zero buffers at the runtime level.
+    unlowerable_effects = {eff for eff in jaxpr.effects
+                           if eff not in lowerable_effects}
+    if unlowerable_effects:
+      raise ValueError(
+          f'Cannot lower jaxpr with unlowerable effects: {unlowerable_effects}')
+    effects = [eff for eff in jaxpr.effects if eff in core.ordered_effects]
     lower_jaxpr_to_fun(
-        ctx, "main", jaxpr, public=True, replace_units_with_dummy=True,
+        ctx, "main", jaxpr, effects, public=True, create_tokens=True,
+        replace_units_with_dummy=True,
         replace_tokens_with_dummy=True, replicated_args=replicated_args,
         arg_shardings=arg_shardings, result_shardings=result_shardings,
         input_output_aliases=input_output_aliases)
@@ -558,11 +567,69 @@ def _set_up_aliases(avals_in, avals_out, donated_args):
 
   return input_output_aliases, out_donated_args
 
+Token = Sequence[ir.Value]
+
+def token_type() -> Sequence[ir.Type]:
+  return [mhlo.TokenType.get()]
+
+def token() -> Token:
+  return wrap_singleton_ir_values(
+      mhlo.CreateTokenOp(mhlo.TokenType.get()).result)
+
+class TokenSet:
+  """An immutable container of tokens to be used to lower effectful jaxprs. When lowering
+  effectful jaxprs, we need to thread MHLO tokens to sequence them. Each effect
+  will need its own token that will be threaded in and out of the effectful
+  primitives. A `TokenSet` encapsulates a set of MHLO tokens that will be
+  used by the lowering rules.
+  """
+  _tokens: typing.OrderedDict[core.Effect, Token]
+
+  def __init__(self, *args, **kwargs):
+    self._tokens = collections.OrderedDict(*args, **kwargs)
+
+  def __len__(self):
+    return len(self._tokens)
+
+  def get(self, effect: core.Effect) -> Token:
+    return self._tokens[effect]
+
+  @classmethod
+  def create(cls, effects: Sequence[core.Effect]) -> TokenSet:
+    """Creates a `TokenSet` corresponding to a list of `core.Effect`s."""
+    tokens = [token() for _ in effects]
+    return TokenSet(zip(effects, tokens))
+
+  def items(self) -> Sequence[Tuple[core.Effect, Token]]:
+    return tuple(self._tokens.items())
+
+  def effects(self) -> Sequence[core.Effect]:
+    return tuple(self._tokens.keys())
+
+  def tokens(self) -> Sequence[Token]:
+    return tuple(self._tokens.values())
+
+  def subset(self, effects: Sequence[core.Effect]) -> TokenSet:
+    """Return a subset of the `TokenSet` restricted to a set of `core.Effect`s."""
+    return TokenSet((eff, self._tokens[eff]) for eff in effects)
+
+  def update_tokens(self, tokens: TokenSet) -> TokenSet:
+    """Returns a new `TokenSet` with tokens replaced with ones from the input `TokenSet`."""
+    new_tokens = []
+    for eff in self.effects():
+      if eff in tokens._tokens:
+        new_tokens.append(tokens._tokens[eff])
+      else:
+        new_tokens.append(self._tokens[eff])
+    return TokenSet(zip(self.effects(), new_tokens))
+
 def lower_jaxpr_to_fun(
     ctx: ModuleContext,
     name: str,
     jaxpr: core.ClosedJaxpr,
+    effects: Sequence[core.Effect],
     *,
+    create_tokens: bool = False,
     public: bool = False,
     replace_units_with_dummy: bool = False,
     replace_tokens_with_dummy: bool = False,
@@ -571,7 +638,7 @@ def lower_jaxpr_to_fun(
     result_shardings: Optional[Sequence[Optional[xc.OpSharding]]] = None,
     use_sharding_annotations: bool = True,
     input_output_aliases: Optional[Sequence[Optional[int]]] = None
-) -> FuncOpType:
+) -> func_dialect.FuncOp:
   """Lowers jaxpr and its callees to an IR function.
 
   Assumes that an MLIR context, location, and insertion point are set.
@@ -581,6 +648,8 @@ def lower_jaxpr_to_fun(
     name: the function name. The name will be uniquified by the symbol table,
       so it is ok to use the same name multiple times.
     jaxpr: the jaxpr to lower.
+    effects: a sequence of `core.Effect`s corresponding to an ordering of tokens
+      that will be created in or used by the lowered function.
     public: if true, the function's visibility is set to "public".
     replace_units_with_dummy: if true, unit arguments/return values are
       replaced with bool arrays of size [0].
@@ -607,10 +676,33 @@ def lower_jaxpr_to_fun(
 
   input_types = map(aval_to_types, jaxpr.in_avals)
   output_types = map(aval_to_types, jaxpr.out_avals)
+  if create_tokens:
+    # If we create the tokens they won't be inputs to the MLIR function.
+    num_tokens = 0
+    token_types = []
+  else:
+    # If we aren't creating tokens they will be the initial inputs to the
+    # MLIR function.
+    num_tokens = len(effects)
+    token_types = [token_type() for _ in effects]
+  input_types = [*token_types, *input_types]
+  output_types = [*token_types, *output_types]
+  if input_output_aliases:
+    token_input_output_aliases = [None] * num_tokens
+    input_output_aliases = [*token_input_output_aliases, *input_output_aliases]
+  if arg_shardings:
+    token_shardings = [None] * num_tokens
+    arg_shardings = [*token_shardings, *arg_shardings]
+  if result_shardings:
+    token_shardings = [None] * num_tokens
+    result_shardings = [*token_shardings, *result_shardings]
+  if replicated_args:
+    token_replicated_args = [False] * num_tokens
+    replicated_args = [*token_replicated_args, *replicated_args]
   flat_input_types = util.flatten(input_types)
   flat_output_types = util.flatten(output_types)
   ftype = ir.FunctionType.get(flat_input_types, flat_output_types)
-  func_op = FuncOp(name, ftype, ip=ctx.ip)
+  func_op = func_dialect.FuncOp(name, ftype, ip=ctx.ip)
   func_op.attributes["sym_visibility"] = ir.StringAttr.get(
       "public" if public else "private")
   ctx.symbol_table.insert(func_op)
@@ -675,6 +767,13 @@ def lower_jaxpr_to_fun(
       flat_args = map(wrap_with_sharding_op, flat_args, ir_arg_shardings)
 
     unflattened_args = util.unflatten(flat_args, map(len, input_types))
+    # We separate out the token inputs and the usual inputs. The token inputs
+    # will be passed to `jaxpr_subcomp` separately from the `args`.
+    token_args, unflattened_args = util.split_list(unflattened_args, [num_tokens])
+    if create_tokens:
+      tokens_in = TokenSet.create(effects)
+    else:
+      tokens_in = TokenSet(zip(effects, token_args))
     args: List[List[ir.Value]] = []
     for aval, arg in zip(jaxpr.in_avals, unflattened_args):
       if replace_units_with_dummy and aval is core.abstract_unit:
@@ -684,11 +783,18 @@ def lower_jaxpr_to_fun(
       else:
         args.append(arg)
     callee_name_stack = xla.extend_name_stack(ctx.name_stack,
-                                              xla.wrap_name(name, 'jit'))
-    out_vals = jaxpr_subcomp(ctx.replace(name_stack=callee_name_stack),
-                             jaxpr.jaxpr, map(ir_constants, jaxpr.consts),
-                             *args)
+                                              util.wrap_name(name, 'jit'))
+    out_vals, tokens_out = jaxpr_subcomp(ctx.replace(name_stack=callee_name_stack),
+                                         jaxpr.jaxpr, tokens_in, map(ir_constants, jaxpr.consts),
+                                         *args)
     outs = []
+    if create_tokens:
+      # If we created the tokens in this function, we are done with them and can
+      # ignore `tokens_out`.
+      pass
+    else:
+      for token in tokens_out.tokens():
+        outs.append(token)
     for aval, out in zip(jaxpr.out_avals, out_vals):
       if replace_units_with_dummy and aval is core.abstract_unit:
         outs.append(ir_constants(np.zeros((), np.bool_)))
@@ -706,28 +812,39 @@ def lower_jaxpr_to_fun(
   return func_op
 
 def _emit_lowering_rule_as_fun(lowering_rule,
-                               ctx: LoweringRuleContext) -> FuncOpType:
+                               ctx: LoweringRuleContext) -> func_dialect.FuncOp:
   """Emits the contents of a lowering rule as a private function."""
   input_types = map(aval_to_ir_types, ctx.avals_in)
   output_types = map(aval_to_ir_types, ctx.avals_out)
+  token_types = [token_type() for _ in ctx.tokens_in.items()]
+  input_types = [*token_types, *input_types]
+  output_types = [*token_types, *output_types]
+
   flat_input_types = util.flatten(input_types)
   flat_output_types = util.flatten(output_types)
   ftype = ir.FunctionType.get(flat_input_types, flat_output_types)
   assert ctx.primitive is not None
-  func_op = FuncOp(ctx.primitive.name, ftype, ip=ctx.module_context.ip)
+  func_op = func_dialect.FuncOp(ctx.primitive.name, ftype,
+                                ip=ctx.module_context.ip)
   func_op.attributes["sym_visibility"] = ir.StringAttr.get("private")
   ctx.module_context.symbol_table.insert(func_op)
   entry_block = func_op.add_entry_block()
   with ir.InsertionPoint(entry_block):
     unflattened_args = util.unflatten(entry_block.arguments,
                                       map(len, input_types))
-    outs = lowering_rule(ctx, *_unwrap_singleton_ir_values(unflattened_args))
+    token_args, unflattened_args = util.split_list(unflattened_args, [len(ctx.tokens_in)])
+    sub_ctx = ctx.replace(tokens_in=TokenSet(zip(ctx.tokens_in.effects(), token_args)))
+    outs = lowering_rule(sub_ctx, *_unwrap_singleton_ir_values(unflattened_args))
+    if sub_ctx.tokens_out:
+      outs = [*sub_ctx.tokens_out.tokens(), outs]
     func_dialect.ReturnOp(util.flatten(map(wrap_singleton_ir_values, outs)))
   return func_op
 
 def jaxpr_subcomp(ctx: ModuleContext, jaxpr: core.Jaxpr,
+                  tokens: TokenSet,
                   consts: Sequence[Sequence[ir.Value]],
-                  *args: Sequence[ir.Value]) -> Sequence[Sequence[ir.Value]]:
+                  *args: Sequence[ir.Value]
+                  ) -> Tuple[Sequence[Sequence[ir.Value]], TokenSet]:
   """Lowers a jaxpr into mHLO, inlined into an existing function.
 
   Assumes that an MLIR context, location, and insertion point are set.
@@ -783,11 +900,28 @@ def jaxpr_subcomp(ctx: ModuleContext, jaxpr: core.Jaxpr,
 
       eqn_ctx = (ctx.replace(name_stack=source_info.name_stack) if
           config.jax_experimental_name_stack else ctx)
+      effects = [eff for eff in eqn.effects if eff in core.ordered_effects]
+      tokens_in = tokens.subset(effects)
       rule_ctx = LoweringRuleContext(
           module_context=eqn_ctx, primitive=eqn.primitive,
-          avals_in=map(aval, eqn.invars), avals_out=map(aval, eqn.outvars))
+          avals_in=map(aval, eqn.invars), avals_out=map(aval, eqn.outvars),
+          tokens_in=tokens_in, tokens_out=None)
       ans = rule(rule_ctx, *map(_unwrap_singleton_ir_values, in_nodes),
                  **eqn.params)
+      if effects:
+        # If there were ordered effects in the primitive, there should be output
+        # tokens we need for subsequent ordered effects.
+        tokens_out = rule_ctx.tokens_out
+        if tokens_out is None:
+          raise ValueError(
+              f'Lowering rule for `{eqn.primitive}` needs to set `tokens_out` '
+              f'because it has effects: {eqn.effects}.')
+        if tokens_out.effects() != tokens_in.effects():
+          raise ValueError(
+              f'Lowering rule for `{eqn.primitive}` '
+              'returns incorrect set of output tokens. '
+              f'Expected: {tuple(tokens_in.effects())} vs. Actual: {tuple(tokens_out.effects())}')
+        tokens = tokens.update_tokens(tokens_out)
 
     try:
       out_nodes = tuple(map(wrap_singleton_ir_values, ans))
@@ -799,7 +933,7 @@ def jaxpr_subcomp(ctx: ModuleContext, jaxpr: core.Jaxpr,
     assert all(isinstance(v, ir.Value) for w in out_nodes for v in w), (ans, eqn)
     assert len(ans) == len(eqn.outvars), (ans, eqn)
     map(write, eqn.outvars, out_nodes)
-  return map(read, jaxpr.outvars)
+  return map(read, jaxpr.outvars), tokens
 
 def _ir_consts(consts):
   unique_consts = {id(const): const for const in consts}
@@ -821,39 +955,52 @@ def lower_fun(fun: Callable, multiple_results: bool = True) -> Callable:
     axis_env = ctx.module_context.axis_env
     with core.extend_axis_env_nd(zip(axis_env.names, axis_env.sizes)):
       jaxpr, _, consts = pe.trace_to_jaxpr_dynamic(wrapped_fun, ctx.avals_in)
-    return jaxpr_subcomp(ctx.module_context, jaxpr, _ir_consts(consts),
-                         *map(wrap_singleton_ir_values, args))
+    out, tokens = jaxpr_subcomp(ctx.module_context, jaxpr, ctx.tokens_in, _ir_consts(consts),
+                                *map(wrap_singleton_ir_values, args))
+    ctx.set_tokens_out(tokens)
+    return out
 
   return f_lowered
 
 
 
 def _call_lowering(fn_name, stack_name, call_jaxpr, backend, ctx, avals_in,
-                   avals_out, *args):
+                   avals_out, tokens_in, *args):
   xla.check_backend_matches(backend, ctx.platform)
   output_types = map(aval_to_ir_types, avals_out)
   flat_output_types = util.flatten(output_types)
+  effects = tokens_in.effects()
   symbol_name = lower_jaxpr_to_fun(ctx, fn_name,
-                                   core.ClosedJaxpr(call_jaxpr, ())).name.value
+                                   core.ClosedJaxpr(call_jaxpr, ()),
+                                   effects).name.value
+  args = [*tokens_in.tokens(), *args]
   call = func_dialect.CallOp(flat_output_types,
                              ir.FlatSymbolRefAttr.get(symbol_name),
                              flatten_lowering_ir_args(args))
-  return util.unflatten(call.results, map(len, output_types))
+  out_nodes = util.unflatten(call.results, map(len, output_types))
+  tokens, out_nodes = util.split_list(out_nodes, [len(effects)])
+  tokens_out = tokens_in.update_tokens(TokenSet(zip(effects, tokens)))
+  return out_nodes, tokens_out
 
 def _xla_call_lower(ctx, *args,
                     backend=None, name, call_jaxpr, donated_invars, inline=None,
                     device=None):
   del device, donated_invars, inline  # Ignored.
-  return _call_lowering(name, xla.wrap_name(name, "jit"), call_jaxpr,
-                        backend, ctx.module_context, ctx.avals_in, ctx.avals_out,
-                        *args)
+  out_nodes, tokens = _call_lowering(
+      name, util.wrap_name(name, "jit"), call_jaxpr, backend,
+      ctx.module_context, ctx.avals_in, ctx.avals_out, ctx.tokens_in, *args)
+  ctx.set_tokens_out(tokens)
+  return out_nodes
 
 register_lowering(xla.xla_call_p, _xla_call_lower)
 
 def _named_call_lowering(ctx, *args, name, backend=None,
                          call_jaxpr):
-  return _call_lowering(name, name, call_jaxpr, backend, ctx.module_context,
-                        ctx.avals_in, ctx.avals_out, *args)
+  out_nodes, tokens = _call_lowering(
+      name, name, call_jaxpr, backend, ctx.module_context,
+      ctx.avals_in, ctx.avals_out, ctx.tokens_in, *args)
+  ctx.set_tokens_out(tokens)
+  return out_nodes
 
 register_lowering(core.named_call_p, _named_call_lowering)
 register_lowering(core.call_p, partial(_named_call_lowering, name="core_call"))
@@ -862,8 +1009,12 @@ register_lowering(core.call_p, partial(_named_call_lowering, name="core_call"))
 def full_like_aval(value, aval: core.ShapedArray) -> ir.Value:
   """Returns an IR constant shaped full of `value` shaped like `aval`."""
   zero = ir_constant(np.array(value, aval.dtype))
-  return mhlo.BroadcastOp(aval_to_ir_type(aval), zero,
-                          dense_int_elements(aval.shape)).result
+  if jax._src.lib.mlir_api_version < 9:
+    return mhlo.BroadcastOp(aval_to_ir_type(aval), zero,
+                            dense_int_elements(aval.shape)).result
+  else:
+    return mhlo.BroadcastOp(zero, dense_int_elements(aval.shape)).result
+
 
 def zeros_like_lowering(ctx, x):
   aval, = ctx.avals_in
@@ -880,17 +1031,8 @@ register_lowering(ad_util.stop_gradient_p, lambda ctx, x: [x])
 
 def compare_mhlo(x, y, direction, type):
   """Creates mhlo.CompareOp."""
-  if jax._src.lib.mlir_api_version >= 5:
-    return mhlo.CompareOp(x, y, mhlo.ComparisonDirectionAttr.get(direction),
-                          mhlo.ComparisonTypeAttr.get(type))
-  dims = ir.RankedTensorType(x.type).shape
-  bool_shape = ir.RankedTensorType.get(dims, ir.IntegerType.get_signless(1))
-  if jax._src.lib.mlir_api_version >= 3:
-    return mhlo.CompareOp(bool_shape, x, y,
-                          mhlo.ComparisonDirectionAttr.get(direction),
-                          mhlo.ComparisonTypeAttr.get(type))
-  return mhlo.CompareOp(bool_shape, x, y, ir.StringAttr.get(direction),
-                        ir.StringAttr.get(type))
+  return mhlo.CompareOp(x, y, mhlo.ComparisonDirectionAttr.get(direction),
+                        mhlo.ComparisonTypeAttr.get(type))
 
 def _minmax_mhlo(op, cmp, x, y):
   """Min/max that compares complex values lexicographically as pairs."""
@@ -1000,25 +1142,56 @@ def cache_lowering(f):
   return cached_lowering
 
 
+
+def xla_computation_to_mhlo_module(xla_computation: xc.XlaComputation
+                                  ) -> ir.Module:
+  module_str = xc._xla.mlir.xla_computation_to_mlir_module(xla_computation)
+  return ir.Module.parse(module_str)
+
+def merge_mhlo_modules(dst_module: ir.Module,
+                       sym_name: str,
+                       src_module: ir.Module) -> str:
+  """Returns the name of src_module's main() function, after renaming."""
+  callee_name = None
+  assert dst_module.context == src_module.context
+  dst_symtab = ir.SymbolTable(dst_module.operation)
+
+  n = len(dst_module.body.operations)
+  for op in src_module.body.operations:
+    dst_module.body.append(op)
+  ops = list(dst_module.body.operations)[n:]
+
+  for op in ops:
+    op = typing.cast(func_dialect.FuncOp, op)
+    old_name = op.name.value
+    if op.name.value == "main":
+      dst_symtab.set_symbol_name(op, sym_name)
+      op.attributes["sym_visibility"] = ir.StringAttr.get("private")
+      callee_name = ir.StringAttr(dst_symtab.insert(op)).value
+      new_name = callee_name
+    else:
+      new_name = ir.StringAttr(dst_symtab.insert(op)).value
+
+    # Replace references to the symbol with the new name
+    for other_op in ops:
+      dst_symtab.replace_all_symbol_uses(
+          old_name, new_name, other_op.operation)
+
+
+  assert callee_name is not None
+  return callee_name
+
+
 def xla_fallback_lowering(prim: core.Primitive):
   @cache_lowering
   def fallback(ctx: LoweringRuleContext, *args, **params):
     module_ctx = ctx.module_context
     xla_computation = xla.primitive_subcomputation(
-        module_ctx.platform, module_ctx.axis_env, prim, *ctx.avals_in, **params)
-    submodule_str = xc._xla.mlir.xla_computation_to_mlir_module(xla_computation)
-    submodule = ir.Module.parse(submodule_str)
-    callee_name = None
-    for op in submodule.body.operations:
-      op = typing.cast(FuncOpType, op)
-      module_ctx.module.body.append(op)
-      if op.name.value == "main":
-        op.attributes["sym_name"] = ir.StringAttr.get(f"xla_fallback_{prim.name}")
-        callee_name = ir.StringAttr(module_ctx.symbol_table.insert(op)).value
-        op.attributes["sym_visibility"] = ir.StringAttr.get("private")
-      else:
-        module_ctx.symbol_table.insert(op)
-
+        module_ctx.platform, module_ctx.axis_env, prim, ctx.avals_in,
+        ctx.avals_out, **params)
+    xla_module = xla_computation_to_mhlo_module(xla_computation)
+    callee_name = merge_mhlo_modules(
+        module_ctx.module, f"xla_fallback_{prim.name}", xla_module)
     output_types = map(aval_to_ir_types, ctx.avals_out)
     flat_output_types = util.flatten(output_types)
     output_type = (ir.TupleType.get_tuple(flat_output_types)
@@ -1029,18 +1202,51 @@ def xla_fallback_lowering(prim: core.Primitive):
                                flatten_lowering_ir_args(args)).result
     if not prim.multiple_results:
       return [call]
-    if jax._src.lib.mlir_api_version < 6:
-      flat_results = [mhlo.GetTupleElementOp(typ, call, i32_attr(i)).result
-                      for i, typ in enumerate(flat_output_types)]
-    else:
-      flat_results = [mhlo.GetTupleElementOp(call, i32_attr(i)).result
-                      for i in range(len(flat_output_types))]
+    flat_results = [mhlo.GetTupleElementOp(call, i32_attr(i)).result
+                    for i in range(len(flat_output_types))]
 
     return util.unflatten(flat_results, map(len, output_types))
   return fallback
 
 register_lowering(ad.custom_lin_p, ad._raise_custom_vjp_error_on_jvp)
 
+def emit_python_callback(platform, callback,
+                         operands: List[ir.Value],
+                         operand_avals: List[core.AbstractValue],
+                         result_avals: List[core.AbstractValue],
+                         has_side_effect: bool) -> Tuple[List[ir.Value], Any]:
+  """Creates an MHLO `CustomCallOp` that calls back to the provided function."""
+  if platform != "cpu":
+    raise ValueError('`EmitPythonCallback` not supported in non-CPU backends.')
+  backend = xb.get_backend(platform)
+  result_shapes = util.flatten(
+      [xla.aval_to_xla_shapes(result_aval) for result_aval in result_avals])
+  operand_shapes = util.flatten(
+      [xla.aval_to_xla_shapes(op_aval) for op_aval in operand_avals])
+  callback_descriptor, keepalive = backend.get_emit_python_callback_descriptor(
+      callback, operand_shapes, result_shapes)
+  descriptor_operand = ir_constant(
+      callback_descriptor, canonicalize_types=False)
+  callback_operands = [descriptor_operand, *operands]
+  result_types = util.flatten(
+      [aval_to_ir_types(aval) for aval in result_avals])
+  result_type = ir.TupleType.get_tuple(result_types)
+  result = mhlo.CustomCallOp(
+      [result_type],
+      callback_operands,
+      call_target_name=ir.StringAttr.get("xla_python_cpu_callback"),
+      has_side_effect=ir.BoolAttr.get(has_side_effect),
+      api_version=i32_attr(2),
+      called_computations=ir.ArrayAttr.get([]),
+      backend_config=ir.StringAttr.get(""),
+      operand_layouts=None,
+      result_layouts=None)
+  results = [
+      mhlo.GetTupleElementOp(result, i32_attr(i)).result
+      for i in range(len(result_types))
+  ]
+  return results, keepalive
+
 # Lax ops missing MLIR lowerings.
 # # TODO(b/203775215): these are missing from the cHLO dialect. Either add
 # # them or port them to Python.
@@ -1053,16 +1259,6 @@ register_lowering(ad.custom_lin_p, ad._raise_custom_vjp_error_on_jvp)
 # lax.erf_inv_p,
 # lax.regularized_incomplete_beta_p,
 
-# # CHLO doesn't have complex lowerings of these primitives (b/203718937)
-# lax.acos_p,
-# lax.acosh_p,
-# lax.asin_p,
-# lax.asinh_p,
-# lax.atan_p,
-# lax.atanh_p,
-# lax.cosh_p,
-# lax.tan_p,
-
 # # CHLO doesn't have a legalization for bf16 (b/203774470)
 # lax.erf_p,
 # lax.erfc_p,
@@ -1070,6 +1266,3 @@ register_lowering(ad.custom_lin_p, ad._raise_custom_vjp_error_on_jvp)
 # # Not present in cHLO or mHLO (b/203798239), although we could just emit the
 # # lowered pattern ourselves.
 # lax.top_k_p,
-
-# # TODO(phawkins): implement these lax ops:
-# lax.rng_bit_generator_p,
